import { d } from './dice';
import { tableData } from './tableData';

class TableService {
  private randomTables: RandomTable[] = [];
  private tableByName: {[tableName: string]: RandomTable} = {};
  private tablesByTags: {[tagName: string]: RandomTable[]} = {};
  private tags: Set<string> = new Set();

  constructor() {
    console.time('TableServiceStartup');
    for (let i in tableData) {
      let table = tableData[i];
      let randomTable: RandomTable = new RandomTable(table);
      this.randomTables.push(randomTable);
      this.tableByName[table.name] = randomTable;
      if (table.tags) {
        for (let j in table.tags) {
          let tag: string = table.tags[j];
          if (!this.tablesByTags[tag]) {
            this.tablesByTags[tag] = [];
          }
          this.tags.add(tag);
          this.tablesByTags[tag].push(randomTable);
        }
      }
    }
    console.timeEnd('TableServiceStartup');
  }

  public getTablesWithTag(tag: string) {
    return this.tablesByTags[tag];
  }

  public getTable(name: string) {
    return this.tableByName[name];
  }

  public getTags() {
    return this.tags;
  }

  public searchForTable(search: string) {
    let answer = this.randomTables.filter((table: RandomTable) => {
      return table.name.toLowerCase().indexOf(search.toLowerCase()) !== -1 || table.getTags().join(' ').toLowerCase().indexOf(search.toLowerCase()) !== -1;
    }).map((table: RandomTable) => {
      let score = search.length / table.name.length;
      return {
        table: table,
        score: score
      };
    }).sort((a: {table: RandomTable, score: number}, b: {table: RandomTable, score: number}) => {
      return a.score - b.score;
    }).map((tableScore: {table: RandomTable, score: number}) => {
      return tableScore.table;
    });

    return answer;
  }
}

/*
  From the DM Guide Book
  Example table p.116 in DM Guide
  d12 + d8 | Encounter
  2 | Sunken ship covered in barnacles (25 percent chance that the ship contains treasure; roll randomly on the treasure tables in chapter 7)
  3 | Sunken ship with reef sharks
  4 | Bed of giant oysters
  5 | Underwater steam vent (25 percent chance that the vent is a portal to the Elemental Plane of Fire)
  ...
  20 | Sunken treasure chest (25 percent chance that it contains something of value; roll treasure randomly using the tables in chapter 7)

  A good index of the tables: https://slyflourish.com/random_tables_of_the_dmg.html

  Other good tables
  https://www.dicegeeks.com/category/random-rpg-tables/
*/

class RandomTable {
  public name: string;
  public dice: number[];
  private indexedResults: TableResult[] = [];
  private config: TableConfig;

  constructor(tableConfig: TableConfig) {
    this.name = tableConfig.name;
    this.dice = tableConfig.dice;
    this.constructIndex(tableConfig.possibleResults);
    this.config = tableConfig;
  }

  public roll(input?: number) {
    let testRoll: number = this.getRoll();
    if (!this.isValidInput(testRoll)) {
      console.log(testRoll);
      debugger;
    }
    if (!input || !this.isValidInput(input)) {
      input = this.getRoll();
    }
    return this.indexedResults[input - 1];
  }

  public getPossibleResults() {
    return this.config.possibleResults;
  }

  public getTags() {
    return this.config.tags || [];
  }

  public getNotes() {
    return this.config.notes;
  }

  private getRoll(): number {
    let answer: number = 0;
    for (let i in this.dice) {
      answer = answer + d(this.dice[i]);
    }
    return answer;
  }

  private isValidInput(input: number) {
    return typeof this.indexedResults[input - 1] !== 'undefined';
  }

  private constructIndex(tableResults: TableResult[]) {
    for (let resultIndex in tableResults) {
      let keys: number[] = tableResults[resultIndex].keys;
      for (let r = Math.min(...keys); r <= Math.max(...keys); r++) {
        this.indexedResults.push(tableResults[resultIndex]);
      }
    }
  }
}

interface TableResult {
  keys: number[];
  value: string;
  nextTable?: string;
  description?: string;
}

interface TableConfig {
  name: string;
  notes?: string;
  tags?: string[];
  dice: number[];
  possibleResults: TableResult[];
}

const tableService: TableService = new TableService();

export {
  RandomTable,
  TableResult,
  TableConfig,
  tableService
}
